from rest_framework import viewsets, mixins, status
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from baquara.users.views import GroupViewSet, GroupAdminViewSet

from courses.models import Course, Class
from courses.workspaces.models import Workspace
from courses.workspaces.serializers import (
    WorkspaceSerializer,
    SimpleWorkspaceSerializer,
    WorkspaceGroupSerializer,
    WorkspaceGroupAdminSerializer,
    # WorkspaceClassSerializer,
    ClassSerializer,
)
from courses.course_material.models import CourseMaterial, File as CourseMaterialFile
from courses.course_material.serializers import CourseMaterialSerializer, FilesSerializer


class WorkspaceViewSet(viewsets.ModelViewSet):
    queryset = Workspace.objects.all()
    serializer_class = WorkspaceSerializer
    permission_classes = [IsAuthenticated]

    def get_serializer_class(self):
        if self.request.query_params.get('simple', None):
            self.serializer_class = SimpleWorkspaceSerializer

        return super().get_serializer_class()

    def _get_groups_data(self, instance, request):
        curr_group_set = instance.groups.values_list('id', flat=True)
        new_group_set = request.data.get('groups', [])
        new_group_set = [g['id'] for g in new_group_set]
        return curr_group_set, new_group_set

    def _get_classes_data(self, instance, request):
        curr_class_set = instance.classes.values_list('id', flat=True)
        new_class_set = request.data.get('classes', [])
        new_class_set = [c['id'] for c in new_class_set]
        return curr_class_set, new_class_set

    def _get_set_diff(self, curr, new):
        remove = set(curr) - set(new)
        add = set(new) - set(curr)
        return add,remove

    def update(self, request, *args, **kwargs):
        instance = self.get_object()
        curr_group_set, new_group_set = self._get_groups_data(instance, request)
        g_add, g_remove = self._get_set_diff(curr_group_set, new_group_set)
        for g in g_add:
            instance.groups.add(g)
        for g in g_remove:
            instance.groups.remove(g)

        curr_class_set, new_class_set = self._get_classes_data(instance, request)
        c_add, c_remove = self._get_set_diff(curr_class_set, new_class_set)
        for c in c_add:
            instance.classes.add(c)
        for c in c_remove:
            instance.classes.remove(c)

        return super().update(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        contract = serializer.save()

        curr_group_set, new_group_set = self._get_groups_data(contract, request)
        g_add, g_remove = self._get_set_diff(curr_group_set, new_group_set)
        for g in g_add:
            contract.groups.add(g)
        for g in g_remove:
            contract.groups.remove(g)

        curr_class_set, new_class_set = self._get_classes_data(contract, request)
        c_add, c_remove = self._get_set_diff(curr_class_set, new_class_set)
        for c in c_add:
            contract.classes.add(c)
        for c in c_remove:
            contract.classes.remove(c)

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class ClassViewSet(viewsets.ModelViewSet):
    queryset = Class.objects.all()
    serializer_class = ClassSerializer
    filter_fields = ('course',)


class WorkspaceGroupViewSet(GroupViewSet):
    """
    Override group viewset add contracts.
    """
    def get_serializer_class(self):
        return WorkspaceGroupSerializer


class WorkspaceGroupAdminViewSet(GroupAdminViewSet):
    """
    Override group viewset add contracts.
    """
    def get_serializer_class(self):
        return WorkspaceGroupAdminSerializer
